@@ -73,6 +73,7 @@
 #include <errno.h>
 #include <assert.h>
 #include "nilfs.h"
+#include "realpath.h"
 
 
 inline static int iseol(int c)
@@ -138,29 +139,37 @@ static int has_mntopt(const char *opts, const char *opt)
 #endif	/* LINE_MAX */
 #define PROCMOUNTS	"/proc/mounts"
 
-static int nilfs_find_fs(struct nilfs *nilfs, const char *dev, const char *opt)
+static int nilfs_find_fs(struct nilfs *nilfs, const char *dev, const char *dir,
+			 const char *opt)
 {
 	FILE *fp;
 	char line[LINE_MAX], *mntent[NMNTFLDS];
 	size_t len;
 	int ret, n;
+	char canonical[PATH_MAX + 2];
 
-	if ((fp = fopen(PROCMOUNTS, "r")) == NULL)
+	if (dev && myrealpath(dev, canonical, sizeof(canonical)))
+		dev = canonical;
+
+	fp = fopen(PROCMOUNTS, "r");
+	if (fp == NULL)
 		return -1;
 	ret = -1;
 	while (fgets(line, sizeof(line), fp) != NULL) {
 		n = tokenize(line, mntent, NMNTFLDS);
 		assert(n == NMNTFLDS);
 
-		if (((dev == NULL) || (strcmp(mntent[MNTFLD_FS], dev) == 0)) &&
-		    (strcmp(mntent[MNTFLD_TYPE], NILFS_FSTYPE) == 0) &&
+		if ((dev == NULL || strcmp(mntent[MNTFLD_FS], dev) == 0) &&
+		    (dir == NULL || strcmp(mntent[MNTFLD_DIR], dir) == 0) &&
+		    strcmp(mntent[MNTFLD_TYPE], NILFS_FSTYPE) == 0 &&
 		    has_mntopt(mntent[MNTFLD_OPTS], opt)) {
-			if ((nilfs->n_dev = strdup(mntent[MNTFLD_FS])) == NULL)
+			nilfs->n_dev = strdup(mntent[MNTFLD_FS]);
+			if (nilfs->n_dev == NULL)
 				break;
 			len = strlen(mntent[MNTFLD_DIR]) +
 				strlen(NILFS_IOC) + 2;
-			if ((nilfs->n_ioc =
-			     (char *)malloc(sizeof(char) * len)) == NULL) {
+			nilfs->n_ioc = malloc(sizeof(char) * len);
+			if (nilfs->n_ioc == NULL) {
 				free(nilfs->n_dev);
 				nilfs->n_dev = NULL;
 				break;
@@ -175,22 +184,9 @@ static int nilfs_find_fs(struct nilfs *nilfs, const char *dev, const char *opt)
 	return ret;
 }
 
-static int nilfs_read_sb(struct nilfs *nilfs)
-{
-	if (lseek(nilfs->n_devfd, NILFS_SB_OFFSET_BYTES, SEEK_SET) < 0)
-		return -1;
-	return read(nilfs->n_devfd, &nilfs->n_sb,
-		    sizeof(struct nilfs_super_block));
-}
-
-struct nilfs_super_block *nilfs_get_sb(struct nilfs *nilfs)
-{
-	return &nilfs->n_sb;
-}
-
 size_t nilfs_get_block_size(struct nilfs *nilfs)
 {
-	return 1UL << (le32_to_cpu(nilfs->n_sb.s_log_block_size) +
+	return 1UL << (le32_to_cpu(nilfs->n_sb->s_log_block_size) +
 		       NILFS_SB_BLOCK_SIZE_SHIFT);
 }
 
@@ -201,7 +197,7 @@ int nilfs_opt_set_mmap(struct nilfs *nilfs)
 
 	if ((pagesize = sysconf(_SC_PAGESIZE)) < 0)
 		return -1;
-	segsize = le32_to_cpu(nilfs->n_sb.s_blocks_per_segment) *
+	segsize = le32_to_cpu(nilfs->n_sb->s_blocks_per_segment) *
 		nilfs_get_block_size(nilfs);
 	if (segsize % pagesize != 0)
 		return -1;
@@ -223,37 +219,42 @@ int nilfs_opt_test_mmap(struct nilfs *nilfs)
 /**
  * nilfs_open - create a NILFS object
  * @dev: device
+ * @dir: mount directory
  * @flags: open flags
  */
-struct nilfs *nilfs_open(const char *dev, int flags)
+struct nilfs *nilfs_open(const char *dev, const char *dir, int flags)
 {
 	struct nilfs *nilfs;
 	int oflags;
 
-	if (!(flags & (NILFS_OPEN_RAW |
-		       NILFS_OPEN_RDONLY |
-		       NILFS_OPEN_WRONLY |
-		       NILFS_OPEN_RDWR))) {
+	if (!(flags & (NILFS_OPEN_RAW | NILFS_OPEN_RDONLY |
+		       NILFS_OPEN_WRONLY | NILFS_OPEN_RDWR))) {
 		errno = EINVAL;
 		return NULL;
 	}
 
-	if ((nilfs = (struct nilfs *)malloc(sizeof(struct nilfs))) == NULL)
+	nilfs = malloc(sizeof(*nilfs));
+	if (nilfs == NULL)
 		return NULL;
+
+	nilfs->n_sb = NULL;
 	nilfs->n_devfd = -1;
 	nilfs->n_iocfd = -1;
 	nilfs->n_dev = NULL;
 	nilfs->n_ioc = NULL;
+	nilfs->n_mincno = NILFS_CNO_MIN;
 
 	if (flags & NILFS_OPEN_RAW) {
 		if (dev == NULL) {
-			if (nilfs_find_fs(nilfs, dev, MNTOPT_RW) < 0)
+			if (nilfs_find_fs(nilfs, dev, dir, MNTOPT_RW) < 0)
 				goto out_fd;
 		} else {
-			if ((nilfs->n_dev = strdup(dev)) == NULL)
+			nilfs->n_dev = strdup(dev);
+			if (nilfs->n_dev == NULL)
 				goto out_fd;
 		}
-		if ((nilfs->n_devfd = open(nilfs->n_dev, O_RDONLY)) < 0)
+		nilfs->n_devfd = open(nilfs->n_dev, O_RDONLY);
+		if (nilfs->n_devfd < 0)
 			goto out_fd;
 		if (nilfs_read_sb(nilfs) < 0)
 			goto out_fd;
@@ -261,10 +262,10 @@ struct nilfs *nilfs_open(const char *dev, int flags)
 
 	if (flags &
 	    (NILFS_OPEN_RDONLY | NILFS_OPEN_WRONLY | NILFS_OPEN_RDWR)) {
-		if (nilfs_find_fs(nilfs, dev, MNTOPT_RW) < 0) {
+		if (nilfs_find_fs(nilfs, dev, dir, MNTOPT_RW) < 0) {
 			if (!(flags & NILFS_OPEN_RDONLY))
 				goto out_nilfs;
-			if (nilfs_find_fs(nilfs, dev, MNTOPT_RO) < 0)
+			if (nilfs_find_fs(nilfs, dev, dir, MNTOPT_RO) < 0)
 				goto out_nilfs;
 		}
 		oflags = O_CREAT;
@@ -274,10 +275,10 @@ struct nilfs *nilfs_open(const char *dev, int flags)
 			oflags |= O_WRONLY;
 		else if (flags & NILFS_OPEN_RDWR)
 			oflags |= O_RDWR;
-		if ((nilfs->n_iocfd = open(nilfs->n_ioc, oflags,
-					   S_IRUSR | S_IWUSR |
-					   S_IRGRP | S_IWGRP |
-					   S_IROTH | S_IWOTH)) < 0)
+		nilfs->n_iocfd = open(nilfs->n_ioc, oflags,
+				      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP |
+				      S_IROTH | S_IWOTH);
+		if (nilfs->n_iocfd < 0)
 			goto out_fd;
 	}
 
@@ -294,6 +295,8 @@ struct nilfs *nilfs_open(const char *dev, int flags)
 		free(nilfs->n_dev);
 	if (nilfs->n_ioc != NULL)
 		free(nilfs->n_ioc);
+	if (nilfs->n_sb != NULL)
+		free(nilfs->n_sb);
 
  out_nilfs:
 	free(nilfs);
@@ -314,12 +317,14 @@ void nilfs_close(struct nilfs *nilfs)
 		free(nilfs->n_dev);
 	if (nilfs->n_ioc != NULL)
 		free(nilfs->n_ioc);
+	if (nilfs->n_sb != NULL)
+		free(nilfs->n_sb);
 	free(nilfs);
 }
 
 /**
  * nilfs_get_dev -
- * @nilfs: 
+ * @nilfs:
  */
 const char *nilfs_get_dev(const struct nilfs *nilfs)
 {
@@ -332,7 +337,7 @@ const char *nilfs_get_dev(const struct nilfs *nilfs)
  * @cno:
  * @mode:
  */
-int nilfs_change_cpmode(const struct nilfs *nilfs, nilfs_cno_t cno, int mode)
+int nilfs_change_cpmode(struct nilfs *nilfs, nilfs_cno_t cno, int mode)
 {
 	struct nilfs_cpmode cpmode;
 
@@ -358,7 +363,7 @@ int nilfs_change_cpmode(const struct nilfs *nilfs, nilfs_cno_t cno, int mode)
  * @cpinfo:
  * @nci:
  */
-ssize_t nilfs_get_cpinfo(const struct nilfs *nilfs, nilfs_cno_t cno, int mode,
+ssize_t nilfs_get_cpinfo(struct nilfs *nilfs, nilfs_cno_t cno, int mode,
 			 struct nilfs_cpinfo *cpinfo, size_t nci)
 {
 	struct nilfs_argv argv;
@@ -367,18 +372,26 @@ ssize_t nilfs_get_cpinfo(const struct nilfs *nilfs, nilfs_cno_t cno, int mode,
 		errno = EBADF;
 		return -1;
 	}
-	if ((mode == NILFS_CHECKPOINT) && (cno < NILFS_CNO_MIN)) {
-		errno = EINVAL;
-		return -1;
+	if (mode == NILFS_CHECKPOINT) {
+		if (cno < NILFS_CNO_MIN) {
+			errno = EINVAL;
+			return -1;
+		} else if (cno < nilfs->n_mincno)
+			cno = nilfs->n_mincno;
 	}
 
-	argv.v_base = cpinfo;
+	argv.v_base = (unsigned long)cpinfo;
 	argv.v_nmembs = nci;
 	argv.v_size = sizeof(struct nilfs_cpinfo);
 	argv.v_index = cno;
 	argv.v_flags = mode;
 	if (ioctl(nilfs->n_iocfd, NILFS_IOCTL_GET_CPINFO, &argv) < 0)
 		return -1;
+	if (mode == NILFS_CHECKPOINT && argv.v_nmembs > 0 &&
+	    cno == nilfs->n_mincno) {
+		if (cpinfo[0].ci_cno > nilfs->n_mincno)
+			nilfs->n_mincno = cpinfo[0].ci_cno;
+	}
 	return argv.v_nmembs;
 }
 
@@ -387,7 +400,7 @@ ssize_t nilfs_get_cpinfo(const struct nilfs *nilfs, nilfs_cno_t cno, int mode,
  * @nilfs:
  * @cno:
  */
-int nilfs_delete_checkpoint(const struct nilfs *nilfs, nilfs_cno_t cno)
+int nilfs_delete_checkpoint(struct nilfs *nilfs, nilfs_cno_t cno)
 {
 	if (nilfs->n_iocfd < 0) {
 		errno = EBADF;
@@ -417,7 +430,7 @@ int nilfs_get_cpstat(const struct nilfs *nilfs, struct nilfs_cpstat *cpstat)
  * @si:
  * @nsi:
  */
-ssize_t nilfs_get_suinfo(const struct nilfs *nilfs, nilfs_segnum_t segnum,
+ssize_t nilfs_get_suinfo(const struct nilfs *nilfs, __u64 segnum,
 			 struct nilfs_suinfo *si, size_t nsi)
 {
 	struct nilfs_argv argv;
@@ -427,7 +440,7 @@ ssize_t nilfs_get_suinfo(const struct nilfs *nilfs, nilfs_segnum_t segnum,
 		return -1;
 	}
 
-	argv.v_base = si;
+	argv.v_base = (unsigned long)si;
 	argv.v_nmembs = nsi;
 	argv.v_size = sizeof(struct nilfs_suinfo);
 	argv.v_index = segnum;
@@ -467,7 +480,7 @@ ssize_t nilfs_get_vinfo(const struct nilfs *nilfs,
 		return -1;
 	}
 
-	argv.v_base = vinfo;
+	argv.v_base = (unsigned long)vinfo;
 	argv.v_nmembs = nvi;
 	argv.v_size = sizeof(struct nilfs_vinfo);
 	if (ioctl(nilfs->n_iocfd, NILFS_IOCTL_GET_VINFO, &argv) < 0)
@@ -491,7 +504,7 @@ ssize_t nilfs_get_bdescs(const struct nilfs *nilfs,
 		return -1;
 	}
 
-	argv.v_base = bdescs;
+	argv.v_base = (unsigned long)bdescs;
 	argv.v_nmembs = nbdescs;
 	argv.v_size = sizeof(struct nilfs_bdesc);
 	if (ioctl(nilfs->n_iocfd, NILFS_IOCTL_GET_BDESCS, &argv) < 0)
@@ -513,12 +526,12 @@ ssize_t nilfs_get_bdescs(const struct nilfs *nilfs,
  * @segnums:
  * @nsegs:
  */
-int nilfs_clean_segments(const struct nilfs *nilfs,
+int nilfs_clean_segments(struct nilfs *nilfs,
 			 struct nilfs_vdesc *vdescs, size_t nvdescs,
 			 struct nilfs_period *periods, size_t nperiods,
-			 nilfs_sector_t *vblocknrs, size_t nvblocknrs,
+			 __u64 *vblocknrs, size_t nvblocknrs,
 			 struct nilfs_bdesc *bdescs, size_t nbdescs,
-			 nilfs_segnum_t *segnums, size_t nsegs)
+			 __u64 *segnums, size_t nsegs)
 {
 	struct nilfs_argv argv[5];
 
@@ -527,51 +540,24 @@ int nilfs_clean_segments(const struct nilfs *nilfs,
 		return -1;
 	}
 
-	argv[0].v_base = vdescs;
+	argv[0].v_base = (unsigned long)vdescs;
 	argv[0].v_nmembs = nvdescs;
 	argv[0].v_size = sizeof(struct nilfs_vdesc);
-	argv[1].v_base = periods;
+	argv[1].v_base = (unsigned long)periods;
 	argv[1].v_nmembs = nperiods;
 	argv[1].v_size = sizeof(struct nilfs_period);
-	argv[2].v_base = vblocknrs;
+	argv[2].v_base = (unsigned long)vblocknrs;
 	argv[2].v_nmembs = nvblocknrs;
-	argv[2].v_size = sizeof(nilfs_sector_t);
-	argv[3].v_base = bdescs;
+	argv[2].v_size = sizeof(__u64);
+	argv[3].v_base = (unsigned long)bdescs;
 	argv[3].v_nmembs = nbdescs;
 	argv[3].v_size = sizeof(struct nilfs_bdesc);
-	argv[4].v_base = segnums;
+	argv[4].v_base = (unsigned long)segnums;
 	argv[4].v_nmembs = nsegs;
-	argv[4].v_size = sizeof(nilfs_segnum_t);
+	argv[4].v_size = sizeof(__u64);
 	return ioctl(nilfs->n_iocfd, NILFS_IOCTL_CLEAN_SEGMENTS, argv);
 }
 
-/**
- * nilfs_timedwait -
- * @nilfs:
- * @cond:
- * @timeout:
- */
-int nilfs_timedwait(const struct nilfs *nilfs, int cond,
-		    struct timespec *timeout)
-{
-	struct nilfs_wait_cond wc;
-
-	if (nilfs->n_iocfd < 0) {
-		errno = EBADF;
-		return -1;
-	}
-
-	wc.wc_cond = cond;
-	if (timeout == NULL)
-		wc.wc_flags = 0;
-	else {
-		wc.wc_flags = 1;
-		wc.wc_timeout.tv_sec = timeout->tv_sec;
-		wc.wc_timeout.tv_nsec = timeout->tv_nsec;
-	}
-	return ioctl(nilfs->n_iocfd, NILFS_IOCTL_TIMEDWAIT, &wc);
-}
-
 /**
  * nilfs_sync -
  * @nilfs:
@@ -613,13 +599,13 @@ ssize_t nilfs_get_segment(struct nilfs *nilfs, unsigned long segnum,
 		return -1;
 	}
 
-	if (segnum >= le64_to_cpu(nilfs->n_sb.s_nsegments)) {
+	if (segnum >= le64_to_cpu(nilfs->n_sb->s_nsegments)) {
 		errno = EINVAL;
 		return -1;
 	}
 
-	segsize = le32_to_cpu(nilfs->n_sb.s_blocks_per_segment) *
-		(1UL << (le32_to_cpu(nilfs->n_sb.s_log_block_size) +
+	segsize = le32_to_cpu(nilfs->n_sb->s_blocks_per_segment) *
+		(1UL << (le32_to_cpu(nilfs->n_sb->s_log_block_size) +
 			 NILFS_SB_BLOCK_SIZE_SHIFT));
 	offset = ((off_t)segnum) * segsize;
 
@@ -664,8 +650,8 @@ int nilfs_put_segment(struct nilfs *nilfs, void *segment)
 
 #ifdef HAVE_MUNMAP
 	if (nilfs_opt_test_mmap(nilfs)) {
-		segsize = le32_to_cpu(nilfs->n_sb.s_blocks_per_segment) *
-			(1UL << (le32_to_cpu(nilfs->n_sb.s_log_block_size) +
+		segsize = le32_to_cpu(nilfs->n_sb->s_blocks_per_segment) *
+			(1UL << (le32_to_cpu(nilfs->n_sb->s_log_block_size) +
 				 NILFS_SB_BLOCK_SIZE_SHIFT));
 		return munmap(segment, segsize);
 	} else {
@@ -677,6 +663,20 @@ int nilfs_put_segment(struct nilfs *nilfs, void *segment)
 #endif	/* HAVE_MUNMAP */
 }
 
+__u64 nilfs_get_segment_seqnum(const struct nilfs *nilfs, void *segment,
+			       __u64 segnum)
+{
+	struct nilfs_segment_summary *segsum;
+	unsigned long blkoff;
+
+	blkoff = (segnum == 0) ?
+		le64_to_cpu(nilfs->n_sb->s_first_data_block) : 0;
+
+	segsum = segment + blkoff *
+		(1UL << (le32_to_cpu(nilfs->n_sb->s_log_block_size) +
+			 NILFS_SB_BLOCK_SIZE_SHIFT));
+	return le64_to_cpu(segsum->ss_seq);
+}
 
 /* internal use only */
 
@@ -696,42 +696,40 @@ static int nilfs_psegment_is_valid(const struct nilfs_psegment *pseg)
 			    le32_to_cpu(pseg->p_segsum->ss_sumbytes) - offset);
 }
 
-void nilfs_psegment_init(struct nilfs_psegment *pseg,
-			 nilfs_segnum_t segnum,
-			 void *seg, size_t nblocks,
-			 const struct nilfs *nilfs)
+void nilfs_psegment_init(struct nilfs_psegment *pseg, __u64 segnum,
+			 void *seg, size_t nblocks, const struct nilfs *nilfs)
 {
-	nilfs_blkoff_t blkoff;
+	unsigned long blkoff, nblocks_per_segment;
 
 	blkoff = (segnum == 0) ?
-		le64_to_cpu(nilfs->n_sb.s_first_data_block) : 0;
+		le64_to_cpu(nilfs->n_sb->s_first_data_block) : 0;
+	nblocks_per_segment = le32_to_cpu(nilfs->n_sb->s_blocks_per_segment);
 
-	pseg->p_blksize = 1UL << (le32_to_cpu(nilfs->n_sb.s_log_block_size) +
+	pseg->p_blksize = 1UL << (le32_to_cpu(nilfs->n_sb->s_log_block_size) +
 				  NILFS_SB_BLOCK_SIZE_SHIFT);
 	pseg->p_nblocks = nblocks;
-	pseg->p_nblocks_per_segment =
-		le32_to_cpu(nilfs->n_sb.s_blocks_per_segment);
-	pseg->p_segblocknr = pseg->p_nblocks_per_segment * segnum + blkoff;
-	pseg->p_seed = le32_to_cpu(nilfs->n_sb.s_crc_seed);
+	pseg->p_maxblocks = nblocks_per_segment - blkoff;
+	pseg->p_segblocknr = (sector_t)nblocks_per_segment * segnum + blkoff;
+	pseg->p_seed = le32_to_cpu(nilfs->n_sb->s_crc_seed);
 
-	pseg->p_segsum = (struct nilfs_segment_summary *)(seg +
-			blkoff * pseg->p_blksize);
+	pseg->p_segsum = seg + blkoff * pseg->p_blksize;
 	pseg->p_blocknr = pseg->p_segblocknr;
 }
 
 int nilfs_psegment_is_end(const struct nilfs_psegment *pseg)
 {
-	return (pseg->p_segblocknr + pseg->p_nblocks <= pseg->p_blocknr) ||
-		(pseg->p_segblocknr + pseg->p_nblocks_per_segment - pseg->p_blocknr < NILFS_PSEG_MIN_BLOCKS) ||
+	return pseg->p_blocknr >= pseg->p_segblocknr + pseg->p_nblocks ||
+		pseg->p_segblocknr + pseg->p_maxblocks - pseg->p_blocknr <
+		NILFS_PSEG_MIN_BLOCKS ||
 		!nilfs_psegment_is_valid(pseg);
 }
 
 void nilfs_psegment_next(struct nilfs_psegment *pseg)
 {
-	unsigned long nblocks;	/* unsigned long ??? */
+	unsigned long nblocks;
 
 	nblocks = le32_to_cpu(pseg->p_segsum->ss_nblocks);
-	pseg->p_segsum = (struct nilfs_segment_summary *)((void *)pseg->p_segsum + nblocks * pseg->p_blksize);
+	pseg->p_segsum = (void *)pseg->p_segsum + nblocks * pseg->p_blksize;
 	pseg->p_blocknr += nblocks;
 }
 
@@ -739,22 +737,22 @@ void nilfs_psegment_next(struct nilfs_psegment *pseg)
 void nilfs_file_init(struct nilfs_file *file,
 		     const struct nilfs_psegment *pseg)
 {
-	size_t blksize, rest;
+	size_t blksize, rest, hdrsize;
 
 	file->f_psegment = pseg;
 	blksize = file->f_psegment->p_blksize;
+	hdrsize = le16_to_cpu(pseg->p_segsum->ss_bytes);
 
-	file->f_finfo = (struct nilfs_finfo *)(pseg->p_segsum + 1);
+	file->f_finfo = (void *)pseg->p_segsum + hdrsize;
 	file->f_blocknr = pseg->p_blocknr +
 		(le32_to_cpu(pseg->p_segsum->ss_sumbytes) + blksize - 1) /
 		blksize;
 	file->f_index = 0;
-	file->f_offset = sizeof(struct nilfs_segment_summary);
+	file->f_offset = hdrsize;
 
 	rest = blksize - file->f_offset % blksize;
 	if (sizeof(struct nilfs_finfo) > rest) {
-		file->f_finfo = (struct nilfs_finfo *)((void *)file->f_finfo +
-						       rest);
+		file->f_finfo = (void *)file->f_finfo + rest;
 		file->f_offset += rest;
 	}
 }
@@ -766,25 +764,24 @@ int nilfs_file_is_end(const struct nilfs_file *file)
 }
 
 static size_t nilfs_binfo_total_size(unsigned long offset,
-				     size_t blksize,
-				     size_t bisize,
-				     size_t n)
+				     size_t blksize, size_t bisize, size_t n)
 {
-	if (blksize - offset % blksize >= bisize * n)
+	size_t binfo_per_block, rest = blksize - offset % blksize;
+
+	if (bisize * n <= rest)
 		return bisize * n;
-	else {
-		n -= (blksize - offset % blksize) / bisize;
-		return blksize - offset % blksize +
-			(n / (blksize / bisize)) * blksize +
-			(n % (blksize / bisize)) * bisize;
-	}
+
+	n -= rest / bisize;
+	binfo_per_block = blksize / bisize;
+	return rest + (n / binfo_per_block) * blksize +
+		(n % binfo_per_block) * bisize;
 }
 
 void nilfs_file_next(struct nilfs_file *file)
 {
-	size_t blksize, rest;
+	size_t blksize, rest, delta;
 	size_t dsize, nsize;
-	size_t tdsize, tnsize;
+	unsigned long ndatablk, nblocks;
 
 	blksize = file->f_psegment->p_blksize;
 
@@ -796,26 +793,23 @@ void nilfs_file_next(struct nilfs_file *file)
 		nsize = NILFS_BINFO_DAT_NODE_SIZE;
 	}
 
-	file->f_blocknr += le32_to_cpu(file->f_finfo->fi_nblocks);
+	nblocks = le32_to_cpu(file->f_finfo->fi_nblocks);
+	ndatablk = le32_to_cpu(file->f_finfo->fi_ndatablk);
+
+	delta = sizeof(struct nilfs_finfo);
+	delta += nilfs_binfo_total_size(file->f_offset + delta,
+					blksize, dsize, ndatablk);
+	delta += nilfs_binfo_total_size(file->f_offset + delta,
+					blksize, nsize, nblocks - ndatablk);
+
+	file->f_blocknr += nblocks;
+	file->f_offset += delta;
+	file->f_finfo = (void *)file->f_finfo + delta;
 
-	tdsize = nilfs_binfo_total_size(
-		file->f_offset + sizeof(struct nilfs_finfo),
-		blksize, dsize,
-		le32_to_cpu(file->f_finfo->fi_ndatablk));
-	tnsize = nilfs_binfo_total_size(
-		file->f_offset + sizeof(struct nilfs_finfo) + tdsize,
-		blksize, nsize,
-		le32_to_cpu(file->f_finfo->fi_nblocks) -
-		le32_to_cpu(file->f_finfo->fi_ndatablk));
-	file->f_offset += sizeof(struct nilfs_finfo) + tdsize + tnsize;
-	file->f_finfo = (struct nilfs_finfo *)((void *)file->f_finfo +
-					       sizeof(struct nilfs_finfo) +
-					       tdsize + tnsize);
 	rest = blksize - file->f_offset % blksize;
 	if (sizeof(struct nilfs_finfo) > rest) {
 		file->f_offset += rest;
-		file->f_finfo = (struct nilfs_finfo *)((void *)file->f_finfo +
-						       rest);
+		file->f_finfo = (void *)file->f_finfo + rest;
 	}
 	file->f_index++;
 }
@@ -873,7 +867,3 @@ void nilfs_block_next(struct nilfs_block *blk)
 
 	blk->b_blocknr++;
 }
-
-/* Local Variables:		*/
-/* eval: (c-set-style "linux")	*/
-/* End:				*/
