@@ -76,7 +76,7 @@
 #ifdef _GNU_SOURCE
 #include <getopt.h>
 const static struct option long_option[] = {
-	{"conffile",required_argument, NULL, 'c'},
+	{"conffile", required_argument, NULL, 'c'},
 	{"help", no_argument, NULL, 'h'},
 	/* internal option for mount.nilfs2 only */
 	{"nofork", no_argument, NULL, 'n'},
@@ -110,6 +110,10 @@ static void nilfs_cleanerd_set_log_priority(struct nilfs_cleanerd *cleanerd)
 	setlogmask(LOG_UPTO(cleanerd->c_config.cf_log_priority));
 }
 
+/**
+ * nilfs_cleanerd_config - load configuration file
+ * @cleanerd: cleanerd object
+ */
 static int nilfs_cleanerd_config(struct nilfs_cleanerd *cleanerd)
 {
 	if (nilfs_cldconfig_read(&cleanerd->c_config,
@@ -126,31 +130,61 @@ static int nilfs_cleanerd_config(struct nilfs_cleanerd *cleanerd)
 	return 0;
 }
 
+/**
+ * nilfs_cleanerd_reconfig - reload configuration file
+ * @cleanerd: cleanerd object
+ */
+static int nilfs_cleanerd_reconfig(struct nilfs_cleanerd *cleanerd)
+{
+	struct nilfs_cldconfig *config = &cleanerd->c_config;
+	time_t prev_prot_period = config->cf_protection_period;
+	int ret;
+
+	ret = nilfs_cleanerd_config(cleanerd);
+	if (!ret) {
+		if (config->cf_protection_period > prev_prot_period) {
+			cleanerd->c_protcno = 0;
+			cleanerd->c_prottime = 0;
+		}
+	}
+	return ret;
+}
+
 #ifndef PATH_MAX
 #define PATH_MAX	8192
 #endif	/* PATH_MAX */
 
+/**
+ * nilfs_cleanerd_create - create cleanerd object
+ * @dev: name of the device on which the cleanerd operates
+ * @conffile: pathname of configuration file
+ */
 static struct nilfs_cleanerd *
-nilfs_cleanerd_create(const char *dev, const char *conffile)
+nilfs_cleanerd_create(const char *dev, const char *dir, const char *conffile)
 {
 	struct nilfs_cleanerd *cleanerd;
 
-	if ((cleanerd = (struct nilfs_cleanerd *)malloc(
-		     sizeof(struct nilfs_cleanerd))) == NULL)
+	cleanerd = malloc(sizeof(*cleanerd));
+	if (cleanerd == NULL)
 		return NULL;
-	
-	if ((cleanerd->c_nilfs = nilfs_open(
-		     dev, NILFS_OPEN_RAW | NILFS_OPEN_RDWR)) == NULL)
+
+	cleanerd->c_nilfs = nilfs_open(dev, dir,
+				       NILFS_OPEN_RAW | NILFS_OPEN_RDWR);
+	if (cleanerd->c_nilfs == NULL)
 		goto out_cleanerd;
 
-	if (conffile == NULL)
-		conffile = NILFS_CLEANERD_CONFFILE;
-	if ((cleanerd->c_conffile = strdup(conffile)) == NULL)
+	cleanerd->c_conffile = strdup(conffile ? : NILFS_CLEANERD_CONFFILE);
+	if (cleanerd->c_conffile == NULL)
 		goto out_nilfs;
+
 	if (nilfs_cleanerd_config(cleanerd) < 0)
 		goto out_conffile;
 
 	/* success */
+	cleanerd->c_running = 0;
+	cleanerd->c_protcno = 0; /* means unspecified */
+	cleanerd->c_prottime = 0;
+
 	return cleanerd;
 
 	/* error */
@@ -174,10 +208,7 @@ static void nilfs_cleanerd_destroy(struct nilfs_cleanerd *cleanerd)
 
 static int nilfs_comp_segimp(const void *elem1, const void *elem2)
 {
-	struct nilfs_segimp *segimp1, *segimp2;
-	
-	segimp1 = (struct nilfs_segimp *)elem1;
-	segimp2 = (struct nilfs_segimp *)elem2;
+	const struct nilfs_segimp *segimp1 = elem1, *segimp2 = elem2;
 
 	if (segimp1->si_importance < segimp2->si_importance)
 		return -1;
@@ -187,16 +218,22 @@ static int nilfs_comp_segimp(const void *elem1, const void *elem2)
 	return (segimp1->si_segnum < segimp2->si_segnum) ? -1 : 1;
 }
 
-#define timespec_isset(ts)	((ts)->tv_sec != 0 || (ts)->tv_nsec != 0)
-#define timespec_clear(ts)	((ts)->tv_sec = (ts)->tv_nsec = 0)
-
 #define NILFS_CLEANERD_NSUINFO	512
 
+/**
+ * nilfs_cleanerd_select_segments - select segments to be reclaimed
+ * @cleanerd: cleanerd object
+ * @sustat: status information on segments
+ * @segnums: array of segment numbers to store selected segments
+ * @nsegs: size of the @segnums array
+ * @ts: pointer to a timespec struct to store sleep time
+ * @prottimep: pointer to time to store lower limit of protected period
+ */
 static ssize_t
 nilfs_cleanerd_select_segments(struct nilfs_cleanerd *cleanerd,
 			       struct nilfs_sustat *sustat,
-			       nilfs_segnum_t *segnums, size_t nsegs,
-			       struct timespec *ts)
+			       __u64 *segnums, size_t nsegs,
+			       struct timespec *ts, __u64 *prottimep)
 {
 	struct nilfs *nilfs;
 	struct nilfs_cldconfig *config;
@@ -204,8 +241,8 @@ nilfs_cleanerd_select_segments(struct nilfs_cleanerd *cleanerd,
 	struct nilfs_segimp *sm;
 	struct nilfs_suinfo si[NILFS_CLEANERD_NSUINFO];
 	struct timeval tv;
-	time_t prottime, oldest;
-	nilfs_segnum_t segnum;
+	__u64 prottime, oldest;
+	__u64 segnum;
 	size_t count;
 	ssize_t nssegs, n;
 	unsigned long long imp, thr;
@@ -242,13 +279,12 @@ nilfs_cleanerd_select_segments(struct nilfs_cleanerd *cleanerd,
 		for (i = 0; i < n; i++) {
 			if (nilfs_suinfo_dirty(&si[i]) &&
 			    !nilfs_suinfo_active(&si[i]) &&
-			    !nilfs_suinfo_volatile_active(&si[i]) &&
 			    !nilfs_suinfo_error(&si[i]) &&
 			    ((imp = (*config->cf_selection_policy.p_importance)(&si[i])) < thr)) {
 				if (si[i].sui_lastmod < oldest)
 					oldest = si[i].sui_lastmod;
 				if (si[i].sui_lastmod < prottime) {
-					sm = (struct nilfs_segimp *)nilfs_vector_get_new_element(smv);
+					sm = nilfs_vector_get_new_element(smv);
 					if (sm == NULL) {
 						nssegs = -1;
 						goto out;
@@ -258,21 +294,34 @@ nilfs_cleanerd_select_segments(struct nilfs_cleanerd *cleanerd,
 				}
 			}
 		}
+		if (n == 0) {
+			syslog(LOG_WARNING,
+			       "inconsistent number of segments: %llu "
+			       "(nsegs=%llu)",
+			       (unsigned long long)nilfs_vector_get_size(smv),
+			       (unsigned long long)sustat->ss_nsegs);
+			break;
+		}
 	}
 	nilfs_vector_sort(smv, nilfs_comp_segimp);
 
 	nssegs = (nilfs_vector_get_size(smv) < nsegs) ?
 		nilfs_vector_get_size(smv) : nsegs;
 	for (i = 0; i < nssegs; i++) {
-		sm = (struct nilfs_segimp *)nilfs_vector_get_element(smv, i);
+		sm = nilfs_vector_get_element(smv, i);
 		assert(sm != NULL);
 		segnums[i] = sm->si_segnum;
 	}
 	if (nssegs == 0) {
-		ts->tv_sec = (oldest < tv.tv_sec) ?
-			(oldest - prottime + 1) : 0;
 		ts->tv_nsec = 0;
+		if (oldest < tv.tv_sec)
+			ts->tv_sec = oldest - prottime + 1;
+		else {
+			ts->tv_sec = config->cf_protection_period + 1;
+			cleanerd->c_running = 0;
+		}
 	}
+	*prottimep = prottime;
 
  out:
 	nilfs_vector_destroy(smv);
@@ -281,30 +330,21 @@ nilfs_cleanerd_select_segments(struct nilfs_cleanerd *cleanerd,
 
 static int nilfs_comp_vdesc_blocknr(const void *elem1, const void *elem2)
 {
-	struct nilfs_vdesc *vdesc1, *vdesc2;
-	
-	vdesc1 = (struct nilfs_vdesc *)elem1;
-	vdesc2 = (struct nilfs_vdesc *)elem2;
+	const struct nilfs_vdesc *vdesc1 = elem1, *vdesc2 = elem2;
 
 	return (vdesc1->vd_blocknr < vdesc2->vd_blocknr) ? -1 : 1;
 }
 
 static int nilfs_comp_vdesc_vblocknr(const void *elem1, const void *elem2)
 {
-	struct nilfs_vdesc *vdesc1, *vdesc2;
-
-	vdesc1 = (struct nilfs_vdesc *)elem1;
-	vdesc2 = (struct nilfs_vdesc *)elem2;
+	const struct nilfs_vdesc *vdesc1 = elem1, *vdesc2 = elem2;
 
 	return (vdesc1->vd_vblocknr < vdesc2->vd_vblocknr) ? -1 : 1;
 }
 
 static int nilfs_comp_period(const void *elem1, const void *elem2)
 {
-	struct nilfs_period *period1, *period2;
-
-	period1 = (struct nilfs_period *)elem1;
-	period2 = (struct nilfs_period *)elem2;
+	const struct nilfs_period *period1 = elem1, *period2 = elem2;
 
 	return (period1->p_start < period2->p_start) ? -1 :
 		(period1->p_start == period2->p_start) ? 0 : 1;
@@ -312,10 +352,7 @@ static int nilfs_comp_period(const void *elem1, const void *elem2)
 
 static int nilfs_comp_bdesc(const void *elem1, const void *elem2)
 {
-	struct nilfs_bdesc *bdesc1, *bdesc2;
-
-	bdesc1 = (struct nilfs_bdesc *)elem1;
-	bdesc2 = (struct nilfs_bdesc *)elem2;
+	const struct nilfs_bdesc *bdesc1 = elem1, *bdesc2 = elem2;
 
 	if (bdesc1->bd_ino < bdesc2->bd_ino)
 		return -1;
@@ -335,6 +372,13 @@ static int nilfs_comp_bdesc(const void *elem1, const void *elem2)
 		return 0;
 }
 
+/**
+ * nilfs_cleanerd_acc_blocks_file - collect summary of blocks in a file
+ * @cleanerd: cleanerd object
+ * @file: file object
+ * @vdescv: vector object to store (descriptors of) virtual block numbers
+ * @bdescv: vector object to store (descriptors of) disk block numbers
+ */
 static int nilfs_cleanerd_acc_blocks_file(struct nilfs_cleanerd *cleanerd,
 					  struct nilfs_file *file,
 					  struct nilfs_vector *vdescv,
@@ -350,8 +394,8 @@ static int nilfs_cleanerd_acc_blocks_file(struct nilfs_cleanerd *cleanerd,
 	ino = le64_to_cpu(file->f_finfo->fi_ino);
 	if (nilfs_file_is_super(file)) {
 		nilfs_block_for_each(&blk, file) {
-			if ((bdesc = (struct nilfs_bdesc *)
-			     nilfs_vector_get_new_element(bdescv)) == NULL)
+			bdesc = nilfs_vector_get_new_element(bdescv);
+			if (bdesc == NULL)
 				return -1;
 			bdesc->bd_ino = ino;
 			bdesc->bd_oblocknr = blk.b_blocknr;
@@ -360,7 +404,7 @@ static int nilfs_cleanerd_acc_blocks_file(struct nilfs_cleanerd *cleanerd,
 					le64_to_cpu(*(__le64 *)blk.b_binfo);
 				bdesc->bd_level = 0;
 			} else {
-				binfo = (union nilfs_binfo *)blk.b_binfo;
+				binfo = blk.b_binfo;
 				bdesc->bd_offset =
 					le64_to_cpu(binfo->bi_dat.bi_blkoff);
 				bdesc->bd_level = binfo->bi_dat.bi_level;
@@ -369,14 +413,14 @@ static int nilfs_cleanerd_acc_blocks_file(struct nilfs_cleanerd *cleanerd,
 	} else {
 		cno = le64_to_cpu(file->f_finfo->fi_cno);
 		nilfs_block_for_each(&blk, file) {
-			if ((vdesc = (struct nilfs_vdesc *)
-			     nilfs_vector_get_new_element(vdescv)) == NULL)
+			vdesc = nilfs_vector_get_new_element(vdescv);
+			if (vdesc == NULL)
 				return -1;
 			vdesc->vd_ino = ino;
 			vdesc->vd_cno = cno;
 			vdesc->vd_blocknr = blk.b_blocknr;
 			if (nilfs_block_is_data(&blk)) {
-				binfo = (union nilfs_binfo *)blk.b_binfo;
+				binfo = blk.b_binfo;
 				vdesc->vd_vblocknr =
 					le64_to_cpu(binfo->bi_v.bi_vblocknr);
 				vdesc->vd_offset =
@@ -392,6 +436,13 @@ static int nilfs_cleanerd_acc_blocks_file(struct nilfs_cleanerd *cleanerd,
 	return 0;
 }
 
+/**
+ * nilfs_cleanerd_acc_blocks_psegment - collect summary of blocks in a log
+ * @cleanerd: cleanerd object
+ * @psegment: partial segment object
+ * @vdescv: vector object to store (descriptors of) virtual block numbers
+ * @bdescv: vector object to store (descriptors of) disk block numbers
+ */
 static int nilfs_cleanerd_acc_blocks_psegment(struct nilfs_cleanerd *cleanerd,
 					      struct nilfs_psegment *psegment,
 					      struct nilfs_vector *vdescv,
@@ -407,8 +458,17 @@ static int nilfs_cleanerd_acc_blocks_psegment(struct nilfs_cleanerd *cleanerd,
 	return 0;
 }
 
+/**
+ * nilfs_cleanerd_acc_blocks_segment - collect summary of blocks in a segment
+ * @cleanerd: cleanerd object
+ * @segnum: segment number to be parsed
+ * @segment: start address of segment data
+ * @nblocks: size of valid logs in the segment (per block)
+ * @vdescv: vector object to store (descriptors of) virtual block numbers
+ * @bdescv: vector object to store (descriptors of) disk block numbers
+ */
 static int nilfs_cleanerd_acc_blocks_segment(struct nilfs_cleanerd *cleanerd,
-					     nilfs_segnum_t segnum,
+					     __u64 segnum,
 					     void *segment,
 					     size_t nblocks,
 					     struct nilfs_vector *vdescv,
@@ -425,34 +485,80 @@ static int nilfs_cleanerd_acc_blocks_segment(struct nilfs_cleanerd *cleanerd,
 	return 0;
 }
 
-static int nilfs_cleanerd_acc_blocks(struct nilfs_cleanerd *cleanerd,
-				     nilfs_segnum_t *segnums, size_t nsegs,
-				     struct nilfs_vector *vdescv,
-				     struct nilfs_vector *bdescv)
+/**
+ * nilfs_deselect_segment - deselect a segment
+ * @segnums: array of selected segments
+ * @nsegs: size of @segnums array
+ * @nr: index number for @segnums array to be deselected
+ */
+static ssize_t nilfs_deselect_segment(__u64 *segnums, size_t nsegs, int nr)
+{
+	if (nr >= nsegs || nsegs == 0)
+		return -1;
+	else if (nr < nsegs - 1) {
+		__u64 tn = segnums[nr];
+
+		memmove(&segnums[nr], &segnums[nr + 1],
+			sizeof(__u64) * (nsegs - 1 - nr));
+		segnums[nsegs - 1] = tn;
+	}
+	return nsegs - 1;
+}
+
+#define nilfs_cnt64_ge(a, b)	((__s64)(a) - (__s64)(b) >= 0)
+
+/**
+ * nilfs_cleanerd_acc_blocks - collect summary of blocks contained in segments
+ * @cleanerd: cleanerd object
+ * @sustat: status information on segments
+ * @segnums: array of selected segments
+ * @nsegs: size of @segnums array
+ * @vdescv: vector object to store (descriptors of) virtual block numbers
+ * @bdescv: vector object to store (descriptors of) disk block numbers
+ */
+static ssize_t nilfs_cleanerd_acc_blocks(struct nilfs_cleanerd *cleanerd,
+					 struct nilfs_sustat *sustat,
+					 __u64 *segnums, size_t nsegs,
+					 struct nilfs_vector *vdescv,
+					 struct nilfs_vector *bdescv)
 {
 	struct nilfs_suinfo si;
+	struct nilfs *nilfs = cleanerd->c_nilfs;
 	void *segment;
-	int ret, i;
+	int ret, i = 0;
+	ssize_t n = nsegs;
+	__u64 segseq;
 
-	for (i = 0; i < nsegs; i++) {
-		if (nilfs_get_suinfo(cleanerd->c_nilfs,
-				     segnums[i], &si, 1) < 0)
+	while (i < n) {
+		if (nilfs_get_suinfo(nilfs, segnums[i], &si, 1) < 0)
 			return -1;
-		if (nilfs_get_segment(cleanerd->c_nilfs,
-				      segnums[i], &segment) < 0)
+		if (nilfs_get_segment(nilfs, segnums[i], &segment) < 0)
 			return -1;
-		ret = nilfs_cleanerd_acc_blocks_segment(cleanerd,
-			segnums[i], segment, si.sui_nblocks, vdescv, bdescv);
-		if ((nilfs_put_segment(cleanerd->c_nilfs, segment) < 0) ||
-		    (ret < 0))
+
+		segseq = nilfs_get_segment_seqnum(nilfs, segment, segnums[i]);
+		if (nilfs_cnt64_ge(segseq, sustat->ss_prot_seq)) {
+			n = nilfs_deselect_segment(segnums, n, i);
+			if (nilfs_put_segment(nilfs, segment) < 0)
+				return -1;
+			continue;
+		}
+		ret = nilfs_cleanerd_acc_blocks_segment(
+			cleanerd, segnums[i], segment, si.sui_nblocks,
+			vdescv, bdescv);
+		if (nilfs_put_segment(nilfs, segment) < 0 || ret < 0)
 			return -1;
+		i++;
 	}
-
-	return 0;
+	return n;
 }
 
 #define NILFS_CLEANERD_NVINFO		512
 
+/**
+ * nilfs_cleanerd_get_vdesc - get information on virtual block addresses
+ * @cleanerd: cleanerd object
+ * @vdescv: vector object storing (descriptors of) virtual block numbers
+ */
 static int nilfs_cleanerd_get_vdesc(struct nilfs_cleanerd *cleanerd,
 				    struct nilfs_vector *vdescv)
 {
@@ -468,16 +574,14 @@ static int nilfs_cleanerd_get_vdesc(struct nilfs_cleanerd *cleanerd,
 		     (j < NILFS_CLEANERD_NVINFO) &&
 			     (i + j < nilfs_vector_get_size(vdescv));
 		     j++) {
-			vdesc = (struct nilfs_vdesc *)
-				nilfs_vector_get_element(vdescv, i + j);
+			vdesc = nilfs_vector_get_element(vdescv, i + j);
 			assert(vdesc != NULL);
 			vinfo[j].vi_vblocknr = vdesc->vd_vblocknr;
 		}
 		if ((n = nilfs_get_vinfo(cleanerd->c_nilfs, vinfo, j)) < 0)
 			return -1;
 		for (j = 0; j < n; j++) {
-			vdesc = (struct nilfs_vdesc *)
-				nilfs_vector_get_element(vdescv, i + j);
+			vdesc = nilfs_vector_get_element(vdescv, i + j);
 			assert((vdesc != NULL) &&
 			       (vdesc->vd_vblocknr == vinfo[j].vi_vblocknr));
 			vdesc->vd_period.p_start = vinfo[j].vi_start;
@@ -490,6 +594,11 @@ static int nilfs_cleanerd_get_vdesc(struct nilfs_cleanerd *cleanerd,
 
 #define NILFS_CLEANERD_NCPINFO	512
 
+/**
+ * nilfs_cleanerd_get_snapshot - get checkpoint numbers of snapshots
+ * @cleanerd: cleanerd object
+ * @ssp: pointer to store array of checkpoint numbers which are snapshots
+ */
 static ssize_t
 nilfs_cleanerd_get_snapshot(const struct nilfs_cleanerd *cleanerd,
 			    nilfs_cno_t **ssp)
@@ -498,6 +607,7 @@ nilfs_cleanerd_get_snapshot(const struct nilfs_cleanerd *cleanerd,
 	struct nilfs_cpinfo cpinfo[NILFS_CLEANERD_NCPINFO];
 	nilfs_cno_t cno, *ss;
 	ssize_t n;
+	__u64 nss = 0;
 	int i, j;
 
 	if (nilfs_get_cpstat(cleanerd->c_nilfs, &cpstat) < 0)
@@ -505,8 +615,8 @@ nilfs_cleanerd_get_snapshot(const struct nilfs_cleanerd *cleanerd,
 	if (cpstat.cs_nsss == 0)
 		return 0;
 
-	if ((ss = (nilfs_cno_t *)malloc(
-		     sizeof(nilfs_cno_t) * cpstat.cs_nsss)) == NULL)
+	ss = malloc(sizeof(*ss) * cpstat.cs_nsss);
+	if (ss == NULL)
 		return -1;
 
 	cno = 0;
@@ -517,30 +627,100 @@ nilfs_cleanerd_get_snapshot(const struct nilfs_cleanerd *cleanerd,
 			free(ss);
 			return -1;
 		}
+		if (n == 0)
+			break;
 		for (j = 0; j < n; j++)
 			ss[i + j] = cpinfo[j].ci_cno;
+		nss += n;
 		cno = cpinfo[n - 1].ci_next;
+		if (cno == 0)
+			break;
 	}
-
+	if (cpstat.cs_nsss != nss)
+		syslog(LOG_WARNING, "snapshot count mismatch: %llu != %llu",
+		       (unsigned long long)cpstat.cs_nsss,
+		       (unsigned long long)nss);
 	*ssp = ss;
-	return cpstat.cs_nsss;
+	return nss;
 }
 
+/**
+ * nilfs_cleanerd_update_prottime - update the minimum of protected CNOs
+ * @cleanerd: cleanerd object
+ * @prottime: new protection time
+ */
+static int nilfs_cleanerd_update_prottime(struct nilfs_cleanerd *cleanerd,
+					  __u64 prottime)
+{
+	struct nilfs_cpstat cpstat;
+	struct nilfs_cpinfo cpinfo[NILFS_CLEANERD_NCPINFO];
+	nilfs_cno_t cno;
+	size_t count;
+	ssize_t n;
+	int i;
+
+	if (nilfs_get_cpstat(cleanerd->c_nilfs, &cpstat) < 0)
+		return -1;
+
+	if (cleanerd->c_prottime > prottime) {
+		syslog(LOG_WARNING, "protection time rewinded: "
+		       "old period >= %llu, new period >= %llu",
+		       (unsigned long long)cleanerd->c_prottime,
+		       (unsigned long long)prottime);
+		cleanerd->c_protcno = 0;
+	} else if (cleanerd->c_prottime == prottime)
+		return 0; /* protection time unchanged */
+
+	cno = (cleanerd->c_protcno == 0) ? NILFS_CNO_MIN : cleanerd->c_protcno;
+	while (cno < cpstat.cs_cno) {
+		count = (cpstat.cs_cno - cno < NILFS_CLEANERD_NCPINFO) ?
+			cpstat.cs_cno - cno : NILFS_CLEANERD_NCPINFO;
+		n = nilfs_get_cpinfo(cleanerd->c_nilfs, cno, NILFS_CHECKPOINT,
+				     cpinfo, count);
+		if (n < 0)
+			return -1;
+		if (n == 0)
+			break;
+
+		for (i = 0; i < n; i++) {
+			if (cpinfo[i].ci_create >= prottime) {
+				cleanerd->c_protcno = cpinfo[i].ci_cno;
+				goto out;
+			}
+		}
+		cno = cpinfo[n - 1].ci_cno + 1;
+	}
+	cleanerd->c_protcno = cno;
+ out:
+	cleanerd->c_prottime = prottime;
+	syslog(LOG_DEBUG, "protected checkpoints = [%llu,%llu] "
+	       "(protection period >= %llu)",
+	       (unsigned long long)cleanerd->c_protcno,
+	       (unsigned long long)cpstat.cs_cno,
+	       (unsigned long long)prottime);
+	return 0;
+}
+
+/*
+ * nilfs_vdesc_is_live - judge if a virtual block address is live or dead
+ * @vdesc: descriptor object of the virtual block address
+ * @protect: the minimum of checkpoint numbers to be protected
+ * @ss: checkpoint numbers of snapshots
+ * @n: size of @ss array
+ */
 static int nilfs_vdesc_is_live(const struct nilfs_vdesc *vdesc,
-			       nilfs_cno_t protect,
-			       const nilfs_cno_t *ss,
+			       nilfs_cno_t protect, const nilfs_cno_t *ss,
 			       size_t n)
 {
 	long low, high, index;
 	int s;
 
-	if ((vdesc->vd_period.p_end == NILFS_CNO_MAX) ||
-	    (vdesc->vd_period.p_end > protect))
+	if (vdesc->vd_period.p_end == NILFS_CNO_MAX ||
+	    vdesc->vd_period.p_end > protect)
 		return 1;
 
-	if ((n == 0) ||
-	    (vdesc->vd_period.p_start > ss[n - 1]) ||
-	    (vdesc->vd_period.p_end <= ss[0]))
+	if (n == 0 || vdesc->vd_period.p_start > ss[n - 1] ||
+	    vdesc->vd_period.p_end <= ss[0])
 		return 0;
 
 	low = 0;
@@ -567,6 +747,16 @@ static int nilfs_vdesc_is_live(const struct nilfs_vdesc *vdesc,
 	return ss[index] < vdesc->vd_period.p_end;
 }
 
+/**
+ * nilfs_cleanerd_toss_vdescs - deselect deletable virtual block numbers
+ * @cleanerd: cleanerd object
+ * @vdescv: vector object storing (descriptors of) virtual block numbers
+ * @periodv: vector object to store deletable checkpoint numbers (periods)
+ * @vblocknrv: vector object to store deletable virtual block numbers
+ *
+ * nilfs_cleanerd_toss_vdescs() deselects virtual block numbers of files
+ * other than the DAT file.
+ */
 static int nilfs_cleanerd_toss_vdescs(struct nilfs_cleanerd *cleanerd,
 				      struct nilfs_vector *vdescv,
 				      struct nilfs_vector *periodv,
@@ -574,7 +764,7 @@ static int nilfs_cleanerd_toss_vdescs(struct nilfs_cleanerd *cleanerd,
 {
 	struct nilfs_vdesc *vdesc;
 	struct nilfs_period *periodp;
-	nilfs_sector_t *vblocknrp;
+	__u64 *vblocknrp;
 	nilfs_cno_t *ss;
 	ssize_t n;
 	int i, j, ret;
@@ -585,18 +775,16 @@ static int nilfs_cleanerd_toss_vdescs(struct nilfs_cleanerd *cleanerd,
 
 	for (i = 0; i < nilfs_vector_get_size(vdescv); i++) {
 		for (j = i; j < nilfs_vector_get_size(vdescv); j++) {
-			vdesc = (struct nilfs_vdesc *)
-				nilfs_vector_get_element(vdescv, j);
+			vdesc = nilfs_vector_get_element(vdescv, j);
 			assert(vdesc != NULL);
-			if (nilfs_vdesc_is_live(vdesc,
-						NILFS_CNO_MAX, /* not supported */
+			if (nilfs_vdesc_is_live(vdesc, cleanerd->c_protcno,
 						ss, n)) {
 				break;
 			}
-			if (((periodp = (struct nilfs_period *)
-			      nilfs_vector_get_new_element(periodv)) == NULL) ||
-			    ((vblocknrp = (nilfs_sector_t *)
-			      nilfs_vector_get_new_element(vblocknrv)) == NULL)) {
+			if ((periodp =
+			     nilfs_vector_get_new_element(periodv)) == NULL ||
+			    (vblocknrp =
+			     nilfs_vector_get_new_element(vblocknrv)) == NULL) {
 				ret = -1;
 				goto out;
 			}
@@ -615,6 +803,11 @@ static int nilfs_cleanerd_toss_vdescs(struct nilfs_cleanerd *cleanerd,
 	return ret;
 }
 
+/**
+ * nilfs_cleanerd_unify_period - unify periods of checkpoint numbers
+ * @cleanerd: cleanerd object
+ * @periodv: vector object storing checkpoint numbers
+ */
 static void nilfs_cleanerd_unify_period(struct nilfs_cleanerd *cleanerd,
 					struct nilfs_vector *periodv)
 {
@@ -624,12 +817,10 @@ static void nilfs_cleanerd_unify_period(struct nilfs_cleanerd *cleanerd,
 	nilfs_vector_sort(periodv, nilfs_comp_period);
 
 	for (i = 0; i < nilfs_vector_get_size(periodv); i++) {
-		base = (struct nilfs_period *)
-			nilfs_vector_get_element(periodv, i);
+		base = nilfs_vector_get_element(periodv, i);
 		assert(base != NULL);
 		for (j = i + 1; j < nilfs_vector_get_size(periodv); j++) {
-			target = (struct nilfs_period *)
-				nilfs_vector_get_element(periodv, j);
+			target = nilfs_vector_get_element(periodv, j);
 			assert(target != NULL);
 			if (base->p_end < target->p_start)
 				break;
@@ -645,6 +836,11 @@ static void nilfs_cleanerd_unify_period(struct nilfs_cleanerd *cleanerd,
 
 #define NILFS_CLEANERD_NBDESCS	512
 
+/**
+ * nilfs_cleanerd_get_bdescs - get information on disk block addresses
+ * @cleanerd: cleanerd object
+ * @bdescv: vector object storing (descriptors of) disk block numbers
+ */
 static int nilfs_cleanerd_get_bdescs(struct nilfs_cleanerd *cleanerd,
 				     struct nilfs_vector *bdescv)
 {
@@ -668,11 +864,23 @@ static int nilfs_cleanerd_get_bdescs(struct nilfs_cleanerd *cleanerd,
 	return 0;
 }
 
+/**
+ * nilfs_bdesc_is_live - judge if a disk block address is live or dead
+ * @bdesc: descriptor object of the disk block address
+ */
 static int nilfs_bdesc_is_live(struct nilfs_bdesc *bdesc)
 {
 	return bdesc->bd_oblocknr == bdesc->bd_blocknr;
 }
 
+/**
+ * nilfs_cleanerd_toss_bdescs - deselect deletable disk block numbers
+ * @cleanerd: cleanerd object
+ * @bdescv: vector object storing (descriptors of) disk block numbers
+ *
+ * nilfs_cleanerd_toss_bdescs() deselects disk block numbers of the DAT file
+ * which don't belong to the latest DAT file.
+ */
 static int nilfs_cleanerd_toss_bdescs(struct nilfs_cleanerd *cleanerd,
 				      struct nilfs_vector *bdescv)
 {
@@ -681,8 +889,7 @@ static int nilfs_cleanerd_toss_bdescs(struct nilfs_cleanerd *cleanerd,
 
 	for (i = 0; i < nilfs_vector_get_size(bdescv); i++) {
 		for (j = i; j < nilfs_vector_get_size(bdescv); j++) {
-			bdesc = (struct nilfs_bdesc *)
-				nilfs_vector_get_element(bdescv, j);
+			bdesc = nilfs_vector_get_element(bdescv, j);
 			assert(bdesc != NULL);
 			if (nilfs_bdesc_is_live(bdesc))
 				break;
@@ -693,42 +900,65 @@ static int nilfs_cleanerd_toss_bdescs(struct nilfs_cleanerd *cleanerd,
 	return 0;
 }
 
-static int nilfs_cleanerd_clean_segments(struct nilfs_cleanerd *cleanerd,
-					 nilfs_segnum_t *segnums, size_t nsegs)
+/**
+ * nilfs_cleanerd_clean_segments - reclaim segments
+ * @cleanerd: cleanerd object
+ * @sustat: status information on segments
+ * @segnums: array of segment numbers storing selected segments
+ * @nsegs: size of the @segnums array
+ * @prottime: lower limit of protected period
+ */
+static ssize_t nilfs_cleanerd_clean_segments(struct nilfs_cleanerd *cleanerd,
+					     struct nilfs_sustat *sustat,
+					     __u64 *segnums, size_t nsegs,
+					     __u64 prottime)
 {
 	struct nilfs_vector *vdescv, *bdescv, *periodv, *vblocknrv;
-	int ret;
+	ssize_t n, ret = -1;
 
 	if (nsegs == 0)
 		return 0;
 
-	ret = -1;
 	vdescv = nilfs_vector_create(sizeof(struct nilfs_vdesc));
 	bdescv = nilfs_vector_create(sizeof(struct nilfs_bdesc));
 	periodv = nilfs_vector_create(sizeof(struct nilfs_period));
-	vblocknrv = nilfs_vector_create(sizeof(nilfs_sector_t));
-	if ((vdescv == NULL) || (bdescv == NULL) ||
-	    (periodv == NULL) || (vblocknrv == NULL))
+	vblocknrv = nilfs_vector_create(sizeof(__u64));
+	if (vdescv == NULL || bdescv == NULL || periodv == NULL ||
+	    vblocknrv == NULL)
 		goto out_vec;
 
-	if ((ret = nilfs_cleanerd_acc_blocks(cleanerd, segnums, nsegs,
-					     vdescv, bdescv)) < 0)
+	n = nilfs_cleanerd_acc_blocks(cleanerd, sustat, segnums, nsegs,
+				      vdescv, bdescv);
+	if (n <= 0) {
+		ret = n;
 		goto out_vec;
-	if ((ret = nilfs_cleanerd_get_vdesc(cleanerd, vdescv)) < 0)
+	}
+
+	ret = nilfs_cleanerd_get_vdesc(cleanerd, vdescv);
+	if (ret < 0)
 		goto out_vec;
 
-	if ((ret = nilfs_lock_write(cleanerd->c_nilfs)) < 0)
+	ret = nilfs_lock_write(cleanerd->c_nilfs);
+	if (ret < 0)
 		goto out_vec;
 
-	if ((ret = nilfs_cleanerd_toss_vdescs(
-		     cleanerd, vdescv, periodv, vblocknrv)) < 0)
+	ret = nilfs_cleanerd_update_prottime(cleanerd, prottime);
+	if (ret < 0)
 		goto out_lock;
+
+	ret = nilfs_cleanerd_toss_vdescs(cleanerd, vdescv, periodv, vblocknrv);
+	if (ret < 0)
+		goto out_lock;
+
 	nilfs_vector_sort(vdescv, nilfs_comp_vdesc_blocknr);
 	nilfs_cleanerd_unify_period(cleanerd, periodv);
 
-	if ((ret = nilfs_cleanerd_get_bdescs(cleanerd, bdescv)) < 0)
+	ret = nilfs_cleanerd_get_bdescs(cleanerd, bdescv);
+	if (ret < 0)
 		goto out_lock;
-	if ((ret = nilfs_cleanerd_toss_bdescs(cleanerd, bdescv)) < 0)
+
+	ret = nilfs_cleanerd_toss_bdescs(cleanerd, bdescv);
+	if (ret < 0)
 		goto out_lock;
 
 	ret = nilfs_clean_segments(cleanerd->c_nilfs,
@@ -740,10 +970,11 @@ static int nilfs_cleanerd_clean_segments(struct nilfs_cleanerd *cleanerd,
 				   nilfs_vector_get_size(vblocknrv),
 				   nilfs_vector_get_data(bdescv),
 				   nilfs_vector_get_size(bdescv),
-				   segnums,
-				   nsegs);
+				   segnums, n);
 	if (ret < 0)
 		syslog(LOG_ERR, "%m");
+	else
+		ret = n;
 
  out_lock:
 	if (nilfs_unlock_write(cleanerd->c_nilfs) < 0)
@@ -780,7 +1011,7 @@ static int daemonize(int nochdir, int noclose, int nofork)
 	if (setsid() < 0)
 		return -1;
 
-	//umask(0);
+	/* umask(0); */
 
 	if (!nochdir && (chdir(ROOTDIR) < 0))
 		return -1;
@@ -835,14 +1066,19 @@ do {						\
 	(ts)->tv_nsec = (tv)->tv_usec * 1000;	\
 } while (0)
 
+/**
+ * nilfs_cleanerd_clean_loop - main loop of the cleaner daemon
+ * @cleanerd: cleanerd object
+ */
 static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 {
 	struct nilfs_sustat sustat;
 	struct timeval curr, target, diff;
-	struct timespec timeout, *tsp;
-	nilfs_segnum_t segnums[NILFS_CLDCONFIG_NSEGMENTS_PER_CLEAN_MAX];
+	struct timespec timeout;
+	__u64 prev_nongc_ctime = 0, prottime;
+	__u64 segnums[NILFS_CLDCONFIG_NSEGMENTS_PER_CLEAN_MAX];
 	sigset_t sigset;
-	int cond, i, n;
+	int i, ns, n;
 
 	sigemptyset(&sigset);
 	if (sigprocmask(SIG_SETMASK, &sigset, NULL) < 0) {
@@ -866,7 +1102,8 @@ static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 		syslog(LOG_ERR, "cannot get time: %m");
 		return -1;
 	}
-	target.tv_sec += nilfs_cleanerd->c_config.cf_cleaning_interval;
+	target.tv_sec += cleanerd->c_config.cf_cleaning_interval;
+	cleanerd->c_running = 1;
 
 	while (1) {
 		if (sigprocmask(SIG_BLOCK, &sigset, NULL) < 0) {
@@ -875,7 +1112,7 @@ static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 		}
 
 		if (nilfs_cleanerd_reload_config) {
-			if (nilfs_cleanerd_config(nilfs_cleanerd)) {
+			if (nilfs_cleanerd_reconfig(cleanerd)) {
 				syslog(LOG_ERR, "cannot configure: %m");
 				return -1;
 			}
@@ -883,33 +1120,46 @@ static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 			syslog(LOG_INFO, "configuration file reloaded");
 		}
 
-		if (nilfs_get_sustat(nilfs_cleanerd->c_nilfs, &sustat) < 0) {
+		if (nilfs_get_sustat(cleanerd->c_nilfs, &sustat) < 0) {
 			syslog(LOG_ERR, "cannot get segment usage stat: %m");
 			return -1;
 		}
+		if (sustat.ss_nongc_ctime != prev_nongc_ctime) {
+			cleanerd->c_running = 1;
+			prev_nongc_ctime = sustat.ss_nongc_ctime;
+		}
+		if (!cleanerd->c_running)
+			goto sleep;
+
 		syslog(LOG_DEBUG, "ncleansegs = %llu",
 		       (unsigned long long)sustat.ss_ncleansegs);
 
-		if ((n = nilfs_cleanerd_select_segments(
-			     nilfs_cleanerd, &sustat, segnums,
-			     nilfs_cleanerd->c_config.cf_nsegments_per_clean,
-			     &timeout)) < 0) {
+		ns = nilfs_cleanerd_select_segments(
+			cleanerd, &sustat, segnums,
+			cleanerd->c_config.cf_nsegments_per_clean,
+			&timeout, &prottime);
+		if (ns < 0) {
 			syslog(LOG_ERR, "cannot select segments: %m");
 			return -1;
 		}
 		syslog(LOG_DEBUG, "%d segment%s selected to be cleaned",
-		       n, (n <= 1) ? "" : "s");
+		       ns, (ns <= 1) ? "" : "s");
 
-		cond = 0;
-		if (n > 0) {
-			if (nilfs_cleanerd_clean_segments(
-				    nilfs_cleanerd, segnums, n) < 0) {
+		if (ns > 0) {
+			n = nilfs_cleanerd_clean_segments(cleanerd, &sustat,
+							  segnums, ns,
+							  prottime);
+			if (n < 0) {
 				syslog(LOG_ERR, "cannot clean segments: %m");
 				return -1;
+			} else if (n > 0) {
+				for (i = 0; i < n; i++)
+					syslog(LOG_DEBUG,
+					       "segment %llu cleaned",
+					       (unsigned long long)segnums[i]);
+			} else {
+				syslog(LOG_DEBUG, "no segments cleaned");
 			}
-			for (i = 0; i < n; i++)
-				syslog(LOG_DEBUG, "segment %llu cleaned",
-				       (unsigned long long)segnums[i]);
 
 			if (gettimeofday(&curr, NULL) < 0) {
 				syslog(LOG_ERR, "cannot get current time: %m");
@@ -920,37 +1170,31 @@ static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 			if (!timercmp(&curr, &target, <)) {
 				target = curr;
 				target.tv_sec +=
-					nilfs_cleanerd->c_config.cf_cleaning_interval;
+					cleanerd->c_config.cf_cleaning_interval;
 				syslog(LOG_DEBUG, "adjust interval");
 				continue;
 			}
 			timersub(&target, &curr, &diff);
 			timeval_to_timespec(&diff, &timeout);
 			target.tv_sec +=
-				nilfs_cleanerd->c_config.cf_cleaning_interval;
-			tsp = &timeout;
-		} else {
-			cond |= NILFS_TIMEDWAIT_SEG_WRITE;
-			tsp = timespec_isset(&timeout) ? &timeout : NULL;
+				cleanerd->c_config.cf_cleaning_interval;
 		}
 
+ sleep:
 		if (sigprocmask(SIG_UNBLOCK, &sigset, NULL) < 0) {
 			syslog(LOG_ERR, "cannot set signal mask: %m");
 			return -1;
 		}
 
-		if (tsp != NULL)
-			syslog(LOG_DEBUG, "wait %ld.%09ld",
-			       timeout.tv_sec, timeout.tv_nsec);
-		else
-			syslog(LOG_DEBUG, "wait");
-		if ((nilfs_timedwait(cleanerd->c_nilfs, cond, tsp) < 0) &&
-		    (errno != ETIME) &&
-		    (errno != EINTR)) {
-			syslog(LOG_ERR, "%m");
-			return -1;
+		syslog(LOG_DEBUG, "wait %ld.%09ld",
+		       timeout.tv_sec, timeout.tv_nsec);
+		if (nanosleep(&timeout, NULL) < 0) {
+			if (errno != EINTR) {
+				syslog(LOG_ERR, "cannot sleep: %m");
+				return -1;
+			}
+			cleanerd->c_running = 1;
 		}
-
 		syslog(LOG_DEBUG, "wake up");
 	}
 }
@@ -958,7 +1202,7 @@ static int nilfs_cleanerd_clean_loop(struct nilfs_cleanerd *cleanerd)
 int main(int argc, char *argv[])
 {
 	char *progname, *conffile;
-	const char *dev;
+	const char *dev, *dir;
 	int status, nofork, c;
 #ifdef _GNU_SOURCE
 	int option_index;
@@ -968,8 +1212,10 @@ int main(int argc, char *argv[])
 		strrchr(argv[0], '/') + 1 : argv[0];
 	conffile = NILFS_CLEANERD_CONFFILE;
 	nofork = 0;
-	dev = NULL;
 	status = 0;
+	dev = NULL;
+	dir = NULL;
+
 #ifdef _GNU_SOURCE
 	while ((c = getopt_long(argc, argv, "c:hn",
 				long_option, &option_index)) >= 0) {
@@ -997,6 +1243,9 @@ int main(int argc, char *argv[])
 	if (optind < argc)
 		dev = argv[optind++];
 
+	if (optind < argc)
+		dir = argv[optind++];
+
 	if (daemonize(0, 0, nofork) < 0) {
 		fprintf(stderr, "%s: %s\n", progname, strerror(errno));
 		exit(1);
@@ -1005,7 +1254,8 @@ int main(int argc, char *argv[])
 	openlog(progname, LOG_PID, LOG_DAEMON);
 	syslog(LOG_INFO, "start");
 
-	if ((nilfs_cleanerd = nilfs_cleanerd_create(dev, conffile)) == NULL) {
+	nilfs_cleanerd = nilfs_cleanerd_create(dev, dir, conffile);
+	if (nilfs_cleanerd == NULL) {
 		syslog(LOG_ERR, "cannot create cleanerd on %s", dev);
 		status = 1;
 		goto out;
@@ -1024,7 +1274,3 @@ int main(int argc, char *argv[])
 
 	exit(status);
 }
-
-/* Local Variables:		*/
-/* eval: (c-set-style "linux")	*/
-/* End:				*/
